#include <iostream>

using namespace std;

int main(int argc, char *argv[])
{
    int age;
    int new_age;
    
    age = 10;
    new_age = (10+1);
    cout << 'age' << " " << 'new_age' << endl;
    return 0;
}
