#include <iostream>

using namespace std;

int main(int argc, char *argv[])
{
    char fifth;
    int first;
    char fourth;
    int second;
    char seventh;
    char sixth;
    int third;
    
    first = 1;
    second = 2;
    third = 3;
    fourth = '"';
    fifth = 'h';
    sixth = 'w';
    seventh = '\'';
    cout << 'first' << " " << 'second' << " " << 'third' << endl;
    cout << 1 << " " << 2 << " " << "hello" << endl;
    cout << 1 << " " << 2.1 << " " << 32 << " " << 21 << " " << 100 << " " << true << " " << false << " " << "hello" << endl;
    cout << -1 << " " << -2.1 << " " << -32 << " " << -21 << " " << -100 << " " << true << " " << false << " " << "hello" << endl;
    return 0;
}
