#include <string>

using namespace std;

int main(int argc, char *argv[])
{
    double A;
    int HELLO;
    int Hello;
    int HelloWorld;
    string Hello_World;
    string Hello_World_1;
    int _A;
    int a;
    string a_2;
    string a_3;
    bool a_4;
    bool a_5;
    
    a = 1;
    A = 1.0;
    _A = 16;
    HELLO = 180150000;
    Hello = 8;
    HelloWorld = 5;
    Hello_World = "Hello";
    Hello_World_1 = "Hello' world";
    a_2 = "Hello";
    a_3 = "Hello\" world";
    a_4 = true;
    a_5 = false;
    return 0;
}
