#include <iostream>
#include <string>

using namespace std;

int main(int argc, char *argv[])
{
    string extra;
    string greeting;
    string name;
    
    greeting = "hello";
    name = "world";
    extra = "Yay";
    cout << greeting << " " << name << " " << extra << endl;
    return 0;
}
