#include <iostream>
#include <string>

using namespace std;

int main(int argc, char *argv[])
{
    string greeting;
    string name;
    
    greeting = "hello";
    name = "world";
    cout << "greeting" << " " << "name" << endl;
    return 0;
}
