#include <iostream>

using namespace std;

int main(int argc, char *argv[])
{
    int first;
    int second;
    int third;
    
    first = 3;
    second = 2;
    third = 1;
    cout << 'first' << " " << 'second' << " " << 'third' << endl;
    return 0;
}
